/**
 * 
 */
package pt.digitalis.sampleApp.util.stats;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.image.BufferedImage;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import org.apache.poi.hssf.usermodel.HSSFRichTextString;
import org.apache.poi.hssf.usermodel.HSSFRow;
import org.apache.poi.hssf.usermodel.HSSFSheet;
import org.apache.poi.hssf.usermodel.HSSFWorkbook;
import org.jfree.chart.ChartFactory;
import org.jfree.chart.ChartUtilities;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.axis.CategoryAxis;
import org.jfree.chart.axis.CategoryLabelPositions;
import org.jfree.chart.axis.NumberAxis;
import org.jfree.chart.plot.CategoryPlot;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.chart.renderer.category.LineAndShapeRenderer;
import org.jfree.data.category.DefaultCategoryDataset;

import pt.digitalis.log.ILogWrapper;
import pt.digitalis.sampleApp.util.Utilities;

/**
 * @author Pedro Viegas <a
 *         href="mailto:pviegas@digitalis.pt">pviegas@digitalis.pt</a>
 * @created Aug 8, 2007
 * 
 */
public class StatsGenerator {

	/** Default XAxis count */
	final private int DEFAULT_XCOUNT = 20;

	/**
	 * A list of series to calculate statistcs on...
	 */
	private List<Series> series = new ArrayList<Series>();

	/** Max time scale */
	private long maxTime;

	/** The generated X Axis Values */
	private long[] XAxisValues;

	/** The generated Y Axis Values for each serie */
	private long[][] YAxisValues;

	/** The logger */
	private ILogWrapper logger;

	/**
	 * Constructor
	 * 
	 * @param logger
	 *            the logger to use
	 */
	public StatsGenerator(ILogWrapper logger) {
		this.logger = logger;
	}

	/**
	 * Add a serie to the list
	 * 
	 * @param serie
	 */
	public void addSerie(Series serie) {
		series.add(serie);
	}

	/**
	 * Add a group of values to the list
	 * 
	 * @param name
	 * @param values
	 */
	public void addSerie(String name, long[] values) {
		addSerie(new Series(name, values));
	}

	/**
	 * Reads the series and created new chart data. Overrides previous ones if
	 * they existed
	 * 
	 * @param xCount
	 *            the total of X values desired (precision)
	 */
	public void buildChartData(int xCount) {

		XAxisValues = new long[xCount];
		YAxisValues = new long[series.size()][xCount];

		maxTime = 0;

		// Parse the time values to see witch is the X Axis max value
		for (Series s : series) {
			long[] values = s.getValues();

			for (int i = 0; i < values.length; i++) {
				if (values[i] > maxTime)
					maxTime = values[i];
			}
		}

		// Fill the X AXis values
		long xStep = maxTime / xCount;
		long xValue = 0;

		for (int i = 0; i < xCount; i++) {
			XAxisValues[i] = xValue;
			xValue += xStep;

			// Initialize all series with 0;
			for (int j = 0; j < series.size(); j++)
				YAxisValues[j][i] = 0;
		}

		// Fill the Y Axis Values
		int serieIndex = 0;

		for (Series s : series) {
			long[] values = s.getValues();

			for (int i = 0; i < values.length; i++) {

				boolean found = false;

				// -1 execution time means an error!
				if (values[i] != -1) {

					// For each value, let's see what X value it corresponds to
					for (int j = 0; j < XAxisValues.length; j++) {

						if (!found)
							if ((j == XAxisValues.length - 1)
									|| (values[i] < XAxisValues[j])) {
								// If it's the last X value or it's larger than
								// the current value it is the previous one. All
								// values since the previous will have this
								// added to them

								found = true;

								if (j > 0)
									YAxisValues[serieIndex][j - 1] = YAxisValues[serieIndex][j - 1] + 1;
							}

						if (found)
							YAxisValues[serieIndex][j] = YAxisValues[serieIndex][j] + 1;
					}
				}
			}

			serieIndex++;
		}
	}

	/**
	 * Exports the stats to Excel
	 * 
	 * @param fileName
	 *            the name of the EXCEL file to export to
	 */
	public void exportToExcel(String fileName) {
		HSSFWorkbook wb = new HSSFWorkbook();
		HSSFSheet sheet = wb.createSheet("ISS Performance Execution data");

		logger.debug("Creating Workbook...");

		if (XAxisValues == null || YAxisValues == null)
			buildChartData(DEFAULT_XCOUNT);

		int xCount = XAxisValues.length;

		int seriesIndex = 0;

		for (Series s : series) {

			// Created a new row of data
			// Start at 3rd row (1st is 0) for readability
			HSSFRow row = sheet.createRow((short) seriesIndex + 2);

			// Create the series label cell
			row.createCell((short) 1).setCellValue(
					new HSSFRichTextString(s.getName()));

			for (int i = 0; i < xCount; i++) {

				// Create a cell and put a value in it.
				row.createCell((short) (i + 2)).setCellValue(
						YAxisValues[seriesIndex][i]);
			}

			seriesIndex++;
		}

		// Created the XAxis data row
		// Start at 3rd row (1st is 0) for readability
		HSSFRow row = sheet.createRow((short) seriesIndex + 2);

		for (int i = 0; i < xCount; i++)
			row.createCell((short) (i + 2)).setCellValue(XAxisValues[i]);

		logger.debug("Workbook created...");

		// Write the output to a file
		FileOutputStream fileOut;
		try {
			fileOut = new FileOutputStream(fileName);
			wb.write(fileOut);
			fileOut.close();
			logger.info("Created an export of the Performance data in \""
					+ fileName + "\"...");

		} catch (FileNotFoundException e) {
			logger.error("Excel export aborted: The file \"" + fileName
					+ "\" could not be created.");
			e.printStackTrace();

		} catch (IOException e) {
			logger
					.error("Excel export aborted: There was an error writing to \""
							+ fileName + "\".");
			e.printStackTrace();

		}
	}

	/**
	 * Creates a Line Chart and saves it to disk
	 * 
	 * @param fileName
	 *            the name of the image file to export to
	 */
	public void createChartAsPNG(String fileName) {

		String scale = "(ms)";
		int scaleFactor = 1;

		if (maxTime > 999)
			if ((maxTime / 1000) > 999) {

				// Minutes scale
				scale = "(m)";
				scaleFactor = 1000 * 60;

			} else {

				// Seconds scale
				scale = "(s)";
				scaleFactor = 1000;
			}

		logger.debug("Creating Chart...");

		if (XAxisValues == null || YAxisValues == null)
			buildChartData(DEFAULT_XCOUNT);

		int xCount = XAxisValues.length;
		int seriesIndex = 0;

		// A new dataset
		DefaultCategoryDataset dataset = new DefaultCategoryDataset();

		// Fill the dataset with all series values
		for (Series s : series) {
			for (int i = 0; i < xCount; i++) {

				// Add the value to the dataset to the associated series
				// and X Value
				dataset.addValue(YAxisValues[seriesIndex][i], s.getName(), Long
						.toString(XAxisValues[i] / scaleFactor));
			}

			seriesIndex++;
		}

		JFreeChart chart = ChartFactory.createLineChart(
				"ISS Performance Chart", // chart title
				"Time " + scale, // domain axis label
				"Processes", // range axis label
				dataset, // data
				PlotOrientation.VERTICAL, // orientation
				true, // include legend
				true, // tooltips
				false // urls
				);

		logger.debug("Chart initialized...");
		logger.debug("Customizing chart...");

		chart.setBackgroundPaint(Color.white);

		CategoryPlot plot = (CategoryPlot) chart.getPlot();
		plot.setBackgroundPaint(Color.decode("0xffffe0")); // Light yellow

		// customize the range axis...
		NumberAxis rangeAxis = (NumberAxis) plot.getRangeAxis();
		rangeAxis.setStandardTickUnits(NumberAxis.createIntegerTickUnits());
		rangeAxis.setAutoRangeIncludesZero(true);

		// customize the domain axis...
		if ((maxTime / scaleFactor) > 9) {

			// More than 1 digit will be probably truncated, so we will show
			// them rotated by 90º
			CategoryAxis xAxis = plot.getDomainAxis();
			xAxis.setCategoryLabelPositions(CategoryLabelPositions.DOWN_90);
		}

		// Customize the renderer...
		LineAndShapeRenderer renderer = (LineAndShapeRenderer) plot
				.getRenderer();
		// LineRenderer3D renderer = (LineRenderer3D) plot.getRenderer();
		// renderer.setXOffset(8);

		renderer.setSeriesStroke(0, new BasicStroke(2.0f,
				BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND, 1.0f,
				new float[] { 10.0f, 6.0f }, 0.0f));
		renderer.setSeriesStroke(1, new BasicStroke(2.0f,
				BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND, 1.0f,
				new float[] { 2.0f, 6.0f }, 0.0f));
		renderer.setSeriesStroke(2, new BasicStroke(2.0f,
				BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND));

		logger.debug("Chart configured...");

		logger.debug("Creating chart image...");
		BufferedImage bi = chart.createBufferedImage(550, 400);

		// Write the output to a file
		FileOutputStream fileOut;
		try {
			fileOut = new FileOutputStream(fileName);

			ChartUtilities.writeBufferedImageAsPNG(fileOut, bi);

			fileOut.close();
			logger.info("Created a chart of the Performance data in \""
					+ fileName + "\"...");

		} catch (FileNotFoundException e) {
			logger.error("Image export aborted: The file \"" + fileName
					+ "\" could not be created.");
			e.printStackTrace();

		} catch (IOException e) {
			logger
					.error("Image Chart export aborted: There was an error writing to \""
							+ fileName + "\".");
			e.printStackTrace();

		}
	}

	/**
	 * Prints an ascii art chart to the standard output
	 */
	public void printAsciiChart() {

		String scale = "(ms)";
		int scaleFactor = 1;

		if (XAxisValues == null || YAxisValues == null)
			buildChartData(DEFAULT_XCOUNT);

		if (maxTime > 999)
			if ((maxTime / 60) > 999) {

				// Minutes scale
				scale = "(m)";
				scaleFactor = 1024 * 60;

			} else {

				// Seconds scale
				scale = "(s)";
				scaleFactor = 1024;
			}

		int xCount = XAxisValues.length;
		int seriesIndex = 0;

		if (logger.isDebugEnabled()) {
			logger.debug("MaxTime: " + maxTime);
			logger.debug("Scale: " + scale);
			logger.debug("Scale factor: " + scaleFactor);
			logger.debug("XAxisDump: ");

			for (int i = 0; i < xCount; i++) {
				logger.debug("   | Value #"
						+ Utilities.fillLeft(Long.toString(i), 3, "0")
						+ ": "
						+ Utilities.fillLeft(Long.toString(XAxisValues[i]), 10)
						+ "ms ["
						+ Utilities.fillLeft(Long.toString(XAxisValues[i]
								/ scaleFactor), 5) + scale + "]");
			}
		}

		System.out.println("");
		System.out.println("");

		for (Series s : series) {
			System.out.print(" " + Utilities.fillRight(s.getName(), 40, "...")
					+ " | ");

			for (int i = 0; i < xCount; i++) {
				System.out.print(Utilities.fillLeft(Long
						.toString(YAxisValues[seriesIndex][i]), 5));
			}

			System.out.println("");
			seriesIndex++;
		}

		// Axis
		System.out.println(Utilities.fillRight("", 41)
				+ Utilities.fillRight(" -", 9 + xCount * 5, "---"));

		// Now the X Axix
		System.out.print(Utilities.fillRight("", 44));

		for (int i = 0; i < xCount; i++) {
			System.out.print(Utilities.fillLeft(Long.toString(XAxisValues[i]
					/ scaleFactor), 5));
		}

		System.out.println(" " + scale);
		System.out.println("");
	}
}
